/*
 * $Id: cremer_pople.c 22 2009-06-08 15:21:24Z revilo $
 *
 *                This source code is part of
 *
 *                        g_puckering
 *
 * g_puckering is an analysis-tool for calculating the Cremer-Pople
 * ring puckering parameters within a given GROMACS trajectory file.
 *
 * written 2007 by Oliver Stueker
 *
 * Acknowledgments:
 *  The Analysis tool bases on a template and several functions from:
 *  GROMACS - GROningen MAchine for Chemical Simulations
 *  an open source program to perform molecular dynamics simulations
 *  under GNU General Public License.
 *  Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 *  check out http://www.gromacs.org for more information
 *
 *  The implementation of Cremer-Pople-Puckering Algorithm
 *  was done by Mark J. Forster for his GPL-program mdxvu
 *  http://sourceforge.net/projects/mdxvu/
 *
 * License:
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include "gmx_puckering.h"
#include <math.h>
#include <string.h>
#include <stdio.h>

/*------------------------------------------------------------
function cremer_pople
        compute Cremer-Pople ring puckering parameters
        for 5 and 6 membered rings, return results
        in float f1,f2,f3.

		This functions original code was part of mdxvu,
		written by Mark J Forster, in September 2005
		http://sourceforge.net/projects/mdxvu/
------------------------------------------------------------*/
void cremer_pople(float *coords_x, float *coords_y, float *coords_z, int nring,
	              float *f1,float *f2,float *f3)
	/**
	 *  coords_x,
	 *  coords_y,
	 *  coords_z	are a matrix containg all cartesian coordinates of relevant ring atoms
	 *  nring		determens if it's a five- or six-membered ring
	 *
	 *  f1,f2,f3 are pointers to floats
	 *  used to carry out results theta,phi2,Q
	 */
	{
	int i,j;
	float xnew[7],ynew[7],znew[7];
	float zdisp[7];
	float xsum,ysum,zsum;
	float xmean,ymean,zmean;
	float R1[3],R2[3],n[3];
	float length,cross[3];
	float pi=3.14159;
	float s,c,NN;
	float Q,sum1,sum2,q2,q3,phi2;
	float sq,theta,term1,term2;
	float stheta;

	/* for hexapyranose ring should pick
	   atoms O5,C1,C2,C3,C4,C5
	   for agreement with Cremer & Pople Sucrose
	   pyranoid ring coordinates
	*/
	/* for furanoid ring pick atoms
	   O2', C5', C4', C3' C2'
	*/
	NN=((float) nring);
	/* copy coords to arrays xnew,ynew,znew */
	for (i=1;i<=nring;i++) {
      xnew[i]=coords_x[i-1];
			ynew[i]=coords_y[i-1];
			znew[i]=coords_z[i-1];
//    printf("cp.c: x: %8.4f y: %8.4f z: %8.4f\n", xnew[i], ynew[i], znew[i]);
			};

	/* find mean x,y,z coordinates */
	zsum=0.0;
	ysum=0.0;
	xsum=0.0;

	for (i=1;i<=nring;i++) {
			xsum += xnew[i];
			ysum += ynew[i];
			zsum += znew[i];
			};

	xmean=xsum/NN;
	ymean=ysum/NN;
	zmean=zsum/NN;

	/* now subtract mean -> origin at geometric
	   mean of coordinates of 6 ring atoms */
	for (i=1;i<=nring;i++) {
			xnew[i] -= xmean;
			ynew[i] -= ymean;
			znew[i] -= zmean;
			};


	/* compute R1,R2 vectors */
	R1[0]=0.0;
	R1[1]=0.0;
	R1[2]=0.0;
	R2[0]=0.0;
	R2[1]=0.0;
	R2[2]=0.0;
	for (i=1;i<=nring;i++) {
			s=sin(2*pi*(i-1)/NN);
			c=cos(2*pi*(i-1)/NN);
			R1[0] += xnew[i]*s;
			R1[1] += ynew[i]*s;
			R1[2] += znew[i]*s;
			R2[0] += xnew[i]*c;
			R2[1] += ynew[i]*c;
			R2[2] += znew[i]*c;
			};
	/*
	printf("vector R1 = %f %f %f \n",R1[0],R1[1],R1[2]);
	printf("vector R2 = %f %f %f \n",R2[0],R2[1],R2[2]);
	*/

	/* compute cross product */
	cross[0] = R1[1]*R2[2]-R1[2]*R2[1];
	cross[1] = R1[2]*R2[0]-R1[0]*R2[2];
	cross[2] = R1[0]*R2[1]-R1[1]*R2[0];
	/* compute length */
	length=sqrt(cross[0]*cross[0]+cross[1]*cross[1]+cross[2]*cross[2]);
	/* now compute n vector */
	n[0] = cross[0]/length;
	n[1] = cross[1]/length;
	n[2] = cross[2]/length;
	/*
	printf("n vector = %f %f %f \n",n[0],n[1],n[2]);
	*/

	/* compute z axis displacements */
	/* EQ 11 OF CREMER AND POPLE PAPER */
	for (i=1;i<=nring;i++) {
	//		k=atnum[i];
			zdisp[i]=xnew[i]*n[0]+ynew[i]*n[1]+znew[i]*n[2];
	/*      printf("atom %d %4.4s z coord = %f z disp = %f \n",
					i,atom_name[k],znew[i],zdisp[i]); */
			};

	/* compute total puckering amplitude */
	sum1=0.0;
	for (i=1;i<=nring;i++) sum1 += (zdisp[i]*zdisp[i]);
	Q=sqrt(sum1);
	/* printf("total puckering amplitude = %5.4f \n",Q); */

	/* now compute q1, phi1 using Eq 12,13 of
	   D.Cremer and J.A.Pople,
	   JACS vol 97, 1354 (1975) */
	/* q1 should be = 0 hence no need to evaluate */

	if (nring==5 || nring==6) {
	/* now compute q2, phi2 using Eq 12,13 of
	   D.Cremer and J.A.Pople,
	   JACS vol 97, 1354 (1975) */
	sum1=0.0;
	sum2=0.0;
	sq=sqrt(2.0/NN);
	for (j=1;j<=nring;j++) {
			term1=(zdisp[j]*cos(4*pi*(j-1)/NN));
			term2=(zdisp[j]*sin(4*pi*(j-1)/NN));
			sum1=sum1+term1;
			sum2=sum2-term2;
			};
	sum1 *= sq;
	sum2 *= sq;
	/* printf("sum1,sum2= %f %f \n",sum1,sum2); */
	q2=sqrt(sum1*sum1+sum2*sum2);
	/* printf("q2 = %8.4f \n",q2); */
	/* acos returns value in range 0-180 degrees */
	phi2=acos((double) (sum1/q2));
	/* convert to degrees */
	phi2=phi2*180.0/pi;
	stheta=sum2/q2;
	if (stheta<0.0) phi2=(360.0-phi2);
	/* printf("phi 2 = %8.4f degrees \n",phi2*180/pi); */
	}; /* end if nring */

	if (nring==6) {
	/* compute q3 using eq 14 of Cremer & Pople */
	q3=0.0;
	for (j=1;j<=nring;j++) {
			if ((j % 2)==0) {
					q3 -= zdisp[j];
			} else {
					q3 += zdisp[j];
			}; /* end if */
	}; /* next j */
	q3=q3/sqrt(NN);
	/* printf("q3 = %8.4f \n",q3); */
	theta=acos(q3/Q);
	/* convert to degrees */
	theta=theta*180.0/pi;
	/* printf("theta = %8.4f degrees \n",theta); // */
	}; /* end if nring==6 */

	/* take angle results back to main program */
	*f1=theta;
	*f2=phi2;
	*f3=Q;

}

