/*
 * $Id: g_puckering.c 25 2009-10-20 16:19:09Z revilo $
 *
 *                This source code is part of
 *
 *                        g_puckering
 *
 * g_puckering is an analysis-tool for calculating the Cremer-Pople
 * ring puckering parameters within a given GROMACS trajectory file.
 *
 * written 2007 by Oliver Stueker
 *
 * Acknowledgments:
 *  The Analysis tool bases on a template and several functions from:
 *  GROMACS - GROningen MAchine for Chemical Simulations
 *  an open source program to perform molecular dynamics simulations
 *  under GNU General Public License.
 *  Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 *  check out http://www.gromacs.org for more information
 *
 *  The implementation of Cremer-Pople-Puckering Algorithm
 *  was done by Mark J. Forster for his GPL-program mdxvu
 *  http://sourceforge.net/projects/mdxvu/
 *
 * License:
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/* This line is only for CVS version info */
//static char *SRCID_template_c = "$Id: g_puckering.c 25 2009-10-20 16:19:09Z revilo $";

#include <string.h>
#include <math.h>
#include "gmx_puckering.h"
#include "statutil.h"
#include "typedefs.h"
#include "smalloc.h"
#include "vec.h"
#include "copyrite.h"
#include "tpxio.h"
#include "index.h"
#include "xvgr.h"
//#include "pbc.h"
//#include "rmpbc.h"

int main(int argc,char *argv[])
{
  /**********************
   *                    *
   * INIT Variables     *
   *                    *
   **********************/

  static char *desc[] = {
    "g_puckering computes the Cremer-Pople-Puckering Parameters of Pyranoses and Hexanoses.\n",
    "\n",
    "The indexfile needs to contain atom-sextuples or atom-quintuples in the order:\n",
    "* O5 C1 C2 C3 C4 C5 for Hexanoses or \n",
    "* O2 C5 C4 C3 C2    for Pyranoses\n",
    "as defined by Cremer and Pople in [1].\n",
    "\n",
    "If the number of atoms in the group is not divisible by the ringsize given ",
    "by -i the program will give an Error and exit. The program will also check  ",
    "if the atomnames match to the scheme given above and give a warning if the ",
    "names don't match, which can be suppressed by using the -noname option.\n",
    "\n",
    "Following plots are available:\n",
    "-ot    cp_theta.xvg     Theta vs. Time\n",
    "-op      cp_phi.xvg     Phi   vs. Time\n",
    "-oq        cp_q.xvg     Q     vs. Time\n",
    "-otp      cp_tp.xvg     Theta vs. Phi\n",
    "-odt  cp_dtheta.xvg     Distribution of Theta\n",
    "-odp    cp_dphi.xvg     Distribution of Phi\n",
    "-opol  cp_polar.xvg     Polar-Plot: Theta vs. Phi (experimental)\n",
    "\n",
    "\n",
    "[1]  Cremer, D.; Pople, J. A., General definition of ring puckering coordinates. ",
    "J. Am. Chem. Soc. 1975, 97, (6), 1354-1358.\n",
    "\n",
    "Acknowledgments:\n",
    "  The Analysis tool bases on a template and several functions from:\n",
    "  GROMACS - GROningen MAchine for Chemical Simulations\n",
    "  an open source program to perform molecular dynamics simulations\n",
    "  under GNU General Public License.\n",
    "  Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.\n",
    "  check out http://www.gromacs.org for more information\n",
    "\n",
    "  The implementation of Cremer-Pople-Puckering Algorithm\n",
    "  was done by Mark J. Forster for his GPL-program mdxvu\n",
    "  http://sourceforge.net/projects/mdxvu/"

  };
  //  Cremer & Pople, JACS, 1975, 97(6), p. 1354

  /* Extra arguments - but note how you always get the begin/end
   * options when running the program, without mentioning them here!
   */
  /** for ARGUMENTS **/
  int     ringsize = 6;
  int     maxwarn = 20;
  bool    bVerbose = FALSE;
  bool    bTestMode = FALSE;
  bool    bTheta = 0, bPhi = 0, bQ = 0;
  bool    bThetaPhi = 0, bPolar = 0;
  bool    bThetaDist = 0, bPhiDist = 0;
  bool    bDistCheck=TRUE;
  bool    bNameCheck=TRUE;

  /** for TOPOLOGY **/
  t_topology top;
  char       title[STRLEN];
  /** for TRAJECTORY **/
  t_trxframe fr;
  rvec       *xtop;
  matrix     box;
  int        status;
  int        flags = TRX_READ_X;
  int        frame = 0;

  /** for INDEX **/
  t_atoms     *atoms;
  char        *grpnm;
  atom_id     *index;
  int         nidx = 0, maxIndex = 0;
  // Stuff added for index - MB - END //

  /** for CP Parameters **/
  int         nRings;
  int         i;
  float       cptheta, cpphi, cpQ;
  FILE       *out;
  int         ring, atom;
  char        sWarning[STRLEN];

  struct SFrame *first_Frame;
  struct SFrame *current_Frame;
  struct SFrame *new_Frame;
  struct SFrame *prev_Frame;
  struct SRing *current_Ring;
  struct SRing *new_Ring;
  struct SRing *prev_Ring;

  // Stuff added for cremer_pople - OS  END //

  t_pargs pa[] = {
    { "-i",       FALSE, etINT, {&ringsize},  "Size of Ring" },
    { "-v",       FALSE, etBOOL,{&bVerbose},  "Be loud and noisy" },
    { "-dist",    FALSE, etBOOL,{&bDistCheck},"Warn if distance between neighboring Ringatoms is larger that 0.3 nm." },
    { "-name",    FALSE, etBOOL,{&bNameCheck},"Warn if Atomnames don't match with Definition by Cremer & Pople." },
    { "-maxwarn", FALSE, etINT, {&maxwarn},
      "HIDDENNumber of warnings after which input processing stops" },
    { "-t",       FALSE, etBOOL,{&bTestMode}, "HIDDENUse Test-Modus"}
  };
  t_filenm fnm[] = {
    { efTPS, NULL,    NULL,       ffREAD  },  /* this is for the topology */
    { efTRX, "-f",    NULL,       ffREAD  },  /* and this for the trajectory */
    { efNDX, "-n",    NULL,       ffREAD  },  /* and this for the index-file */
    { efXVG, "-ot",   "cp_theta", ffOPTWR },  /* theta vs. time Plot*/
    { efXVG, "-op",   "cp_phi",   ffOPTWR },  /* phi   vs. time Plot*/
    { efXVG, "-oq",   "cp_q",     ffOPTWR },  /* Q     vs. time Plot*/
    { efXVG, "-otp",  "cp_tp",    ffOPTWR },  /* theta vs. Phi  Plot ???Useful??? */
    { efXVG, "-opol", "cp_polar", ffOPTWR },
    { efXVG, "-odt",  "cp_dtheta",ffOPTWR },
    { efXVG, "-odp",  "cp_dphi",  ffOPTWR }
  };

#define NFILE asize(fnm)
  /**********************
   *                    *
   * Parse Arguments    *
   *                    *
   **********************/

  /*CopyRight(stderr,argv[0]);*/
  static char sCopyright[11][STRLEN] = {
    "\n",
    "                             :-)  g_puckering  (-:\n",
    "\n",
    "                         written 2007 by Oliver Stueker\n",
    "\n",
    "         This program is free software; you can redistribute it and/or\n",
    "          modify it under the terms of the GNU General Public License\n",
    "         as published by the Free Software Foundation; either version 2\n",
    "             of the License, or (at your option) any later version.\n"
    "\n",
    "\n",
  };
  for ( i=0; i<11 ; i++)
  {
    fprintf(stderr, sCopyright[i]);
  }

  /* This is the routine responsible for adding default options,
   * calling the X/motif interface, etc. */
  parse_common_args(&argc,argv,PCA_CAN_TIME | PCA_CAN_VIEW,
           NFILE,fnm,asize(pa),pa,asize(desc),desc,0,NULL);

  init_warning(maxwarn);
  bTheta    = opt2bSet("-ot",NFILE,fnm);
  bPhi      = opt2bSet("-op",NFILE,fnm);
  bQ        = opt2bSet("-oq",NFILE,fnm);
  bThetaPhi = opt2bSet("-otp",NFILE,fnm);
  bPolar    = opt2bSet("-opol",NFILE,fnm);
  bThetaDist= opt2bSet("-odt",NFILE,fnm);
  bPhiDist  = opt2bSet("-odp",NFILE,fnm);

  /**********************
   *                    *
   * Read Topology      *
   *                    *
   **********************/
  read_tps_conf(ftp2fn(efTPS,NFILE,fnm),title,&top,&xtop,NULL,box,TRUE);
  sfree(xtop);

  /**********************
   *                    *
   * Read Index File    *
   *                    *
   **********************/
  atoms=&(top.atoms);
  get_index(atoms,ftp2fn_null(efNDX,NFILE,fnm),1,&nidx,&index,&grpnm);

  /**********************
   *                    *
   * Checking           *
   *                    *
   **********************/
  /* Check if No of Elements matches with number of ring atoms */
  nRings=nidx/ringsize;
  fprintf(stdout, "Will be checking %3i Rings.\n", nRings);
  if ((nidx % ringsize) != 0)
    gmx_fatal(FARGS,"number of index elements not multiple of %d, "
    "these can not be %s\n",
    ringsize,(ringsize==6) ? "six-membered rings" : "five-membered rings");
  else {
    if (bVerbose) fprintf(stdout, "OK! Number of index elements matches.\n" );
  }


  /**********************
   *                    *
   * Setup Data Arrays  *
   *                    *
   **********************/
  fprintf(stdout, "Number of Rings: %i\n", nRings);
  int ringAtoms[nRings][ringsize]; // MUSS BLEIBEN!!!!

  /**********************
   *                    *
   * Checking           *
   *                    *
   **********************/
  /* Check if all atoms in Index-File are found in Topology */
  if (bVerbose) fprintf(stdout, "Index consists of items:" );
  for(i=0; i<nidx; i++) {
    ringAtoms[i/ringsize][i%ringsize] = index[i] ;
    if (bVerbose) {
      if (! i%ringsize) fprintf(stdout, "\n");
      fprintf(stdout, " %6i", index[i] );
    }
    if ( index[i] >  maxIndex)
      maxIndex = index[i];
  }
  if (bVerbose) fprintf(stdout, "\n" );
  sfree(index);
  if(maxIndex<0 || maxIndex>(top.atoms.nr))
  {
    gmx_fatal(FARGS,"Error: Atom number %d is out of range.\n",maxIndex);
  }
  if (bVerbose) fprintf(stdout, "OK! Index doesen't exeed Topology.\n" );

  if (bNameCheck) {
    /* Check if Atomnames match
     * to hexapyranose ring (O5,C1,C2,C3,C4,C5)
     * or furanoid ring (O2', C5', C4', C3' C2')
     **/
    for (ring = 0; ring < nRings; ++ring) {
      if (ringsize==6) {
        //top.atoms.atomname[ringAtoms[ring][atom]]
        if (  (
            strcasecmp( *(top.atoms.atomname[ringAtoms[ring][0]]), "O5" ) ||
            strcasecmp( *(top.atoms.atomname[ringAtoms[ring][1]]), "C1" ) ||
            strcasecmp( *(top.atoms.atomname[ringAtoms[ring][2]]), "C2" ) ||
            strcasecmp( *(top.atoms.atomname[ringAtoms[ring][3]]), "C3" ) ||
            strcasecmp( *(top.atoms.atomname[ringAtoms[ring][4]]), "C4" ) ||
            strcasecmp( *(top.atoms.atomname[ringAtoms[ring][5]]), "C5" ) )  )
          {
            sprintf(sWarning, "Names of Atoms specified in Index Group do not match specification for hexapyranose rings.\nIs: (%s,%s,%s,%s,%s,%s), should be: (O5,C1,C2,C3,C4,C5)",
                *(top.atoms.atomname[ringAtoms[ring][0]]),
                *(top.atoms.atomname[ringAtoms[ring][1]]),
                *(top.atoms.atomname[ringAtoms[ring][2]]),
                *(top.atoms.atomname[ringAtoms[ring][3]]),
                *(top.atoms.atomname[ringAtoms[ring][4]]),
                *(top.atoms.atomname[ringAtoms[ring][5]])
            );
            warning(sWarning);
          }
      } else if ( !(
              strcasecmp( *(top.atoms.atomname[ringAtoms[ring][0]]), "O2" ) ||
              strcasecmp( *(top.atoms.atomname[ringAtoms[ring][1]]), "C5" ) ||
              strcasecmp( *(top.atoms.atomname[ringAtoms[ring][2]]), "C4" ) ||
              strcasecmp( *(top.atoms.atomname[ringAtoms[ring][3]]), "C3" ) ||
              strcasecmp( *(top.atoms.atomname[ringAtoms[ring][4]]), "C2" ) ) )
      {
        sprintf(sWarning, "Names of Atoms specified in Index Group do not match specification for furanoid rings.\nIs: (%s,%s,%s,%s,%s), should be: (O2,C5,C4,C3,C2)",
            *(top.atoms.atomname[ringAtoms[ring][0]]),
            *(top.atoms.atomname[ringAtoms[ring][1]]),
            *(top.atoms.atomname[ringAtoms[ring][2]]),
            *(top.atoms.atomname[ringAtoms[ring][3]]),
            *(top.atoms.atomname[ringAtoms[ring][4]])
        );
        warning(sWarning);
      }
    }
  } /*end if (bNameCheck)*/

  frame = 0;
  /* The first time we read data is a little special */
  if (bVerbose) fprintf(stdout, "Now we'll Read the first Frame:\n" );
  read_first_frame(&status,ftp2fn(efTRX,NFILE,fnm),&fr,flags);
  printf("\n");

  first_Frame=(struct SFrame *)malloc(sizeof(struct SFrame));
  if (first_Frame == NULL) {
    gmx_mem("Cannot allocate SFrame.");
    return 1;
  }
  current_Frame=first_Frame;
  prev_Frame=NULL;
  /* This is the main loop over frames */
  //if (!bTestMode)
  current_Frame->next=NULL;
  do {
    /* TODO: remove PBC
     *
     * file://gmx/src/tools/gmc_sdf.c:336
     *
     * #include "pbc.h"
     * #include "rmpbc.h"
     * // Must init pbc every step because of pressure coupling //
     * set_pbc(&pbc,box);
     * rm_pbc(&top.idef,natoms,box,x,x);
     *
     */
    ring = atom = 0;
    current_Frame->fTime=fr.time;
    current_Frame->nNumber=frame;
    current_Frame->rings=(struct SRing *)malloc(sizeof(struct SRing));
    if (current_Frame->rings == NULL) {
      gmx_mem("Cannot allocate SRing.");
      return 1;
    }
    current_Ring=current_Frame->rings;
    prev_Ring=NULL;
    for (ring = 0; ring < nRings; ring++) { /* Loop over rings in frame */
      current_Ring->fTheta = 0.0;
      current_Ring->fPhi   = 0.0;
      current_Ring->fQ     = 0.0;
      for (atom = 0; atom < ringsize; atom++) {
        /* copy coordinates of all ring atoms into the Data Structure */
        current_Ring->aXcoords[atom] = fr.x[ringAtoms[ring][atom]][XX] ;
        current_Ring->aYcoords[atom] = fr.x[ringAtoms[ring][atom]][YY] ;
        current_Ring->aZcoords[atom] = fr.x[ringAtoms[ring][atom]][ZZ] ;

        /* printf("g_p.c: x: %8.4f y: %8.4f z: %8.4f\n", coords_x[ring], coords_y[ring], coords_z[ring]); */
        if (0 && bVerbose) {
          fprintf(stdout, "Atom name: %s\t",*(top.atoms.atomname[ringAtoms[ring][atom]]));
          fprintf(stdout, "Atom charge: %f\n",top.atoms.atom[ringAtoms[ring][atom]].q);
        }
      } //for (atom = 0; atom < ringsize; atom++)

      if (bDistCheck) {
        /* Distance Check: Warn if distance between two neighboring Ring-Atoms is > 0.3nm.  */
        for (atom = 0; atom < ringsize; atom++) {
          float dist=0.0;
          float dX, dY, dZ;
          dX= current_Ring->aXcoords[atom] - current_Ring->aXcoords[(atom+1)%ringsize];
          dY= current_Ring->aYcoords[atom] - current_Ring->aYcoords[(atom+1)%ringsize];
          dZ= current_Ring->aZcoords[atom] - current_Ring->aZcoords[(atom+1)%ringsize];

          dist = sqrt( (dX * dX) + (dY * dY) + (dZ * dZ) );

          if (dist>0.3) {
            sprintf( sWarning, "Distance between Atoms %4i (%s) and %4i (%s) is very large (%5.3G nm).",
                ringAtoms[ring][atom]+1,              *(top.atoms.atomname[ringAtoms[ring][atom]]),
                ringAtoms[ring][(atom+1)%ringsize]+1, *(top.atoms.atomname[ringAtoms[ring][(atom+1)%ringsize]]),
                dist
                );
            warning(sWarning);
            printf("X1: %8.4f, X2: %8.4f, => dX: %f\n", current_Ring->aXcoords[atom], current_Ring->aXcoords[(atom+1)%ringsize], dX);
            printf("Y1: %8.4f, Y2: %8.4f, => dY: %f\n", current_Ring->aYcoords[atom], current_Ring->aYcoords[(atom+1)%ringsize], dY);
            printf("Z1: %8.4f, Z2: %8.4f, => dZ: %f\n", current_Ring->aZcoords[atom], current_Ring->aZcoords[(atom+1)%ringsize], dZ);
          }
        }
      }// end if (bDistanceCheck);

      cremer_pople(current_Ring->aXcoords, current_Ring->aYcoords, current_Ring->aZcoords, ringsize,
                  &cptheta, &cpphi, &cpQ);
      current_Ring->fTheta = cptheta;
      current_Ring->fPhi   = cpphi;
      current_Ring->fQ     = cpQ;
      if ( /*bVerbose ||*/ bTestMode) {
        fprintf(stdout, "CREMER POPLE theta  = %8.2f \n",cptheta);
        fprintf(stdout, "CREMER POPLE phi2   = %8.2f \n",cpphi);
        fprintf(stdout, "CREMER POPLE Q      = %8.5f \n",cpQ);
      }

      /* Append a new Ring to List */
      new_Ring = (struct SRing *)malloc(sizeof(struct SRing));
      if (new_Ring == NULL) {
        gmx_mem("Cannot allocate SRing.");
        return 1;
      }
      current_Ring->next=new_Ring;
      prev_Ring=current_Ring;
      current_Ring=new_Ring;
    }
    /* Terminate List and clean up */
    prev_Ring->next=NULL;
    free(current_Ring);
    current_Ring=NULL;

    /* Appennd new Frame to the List */
    new_Frame = (struct SFrame *)malloc(sizeof(struct SFrame));
    if (new_Frame == NULL) {
      gmx_mem("Cannot allocate SFrame.");
      return 1;
    }
    current_Frame->next=new_Frame;
    prev_Frame=current_Frame;
    current_Frame=new_Frame;
    frame++;
  } while(read_next_frame(status,&fr));
  /* Terminate List and clean up */
  prev_Frame->next=NULL;
  free(current_Frame);
  current_Frame=NULL;

  if (bTestMode) {
    selfTest();
  }


  /**********************
   *                    *
   * Output XVG FILES   *
   *                    *
   **********************/
  fprintf(stdout, "Ready for Output:\n" );
  if (bTheta) {
  /* Write ThethaXVG */
    if (ringsize==5) {
      char sWarn[STRLEN] ;
      sprintf(sWarn, "Puckering Angle Theta is not defined for five-membered rings.\nOutput File %s will not be written.\n", opt2fn("-ot",NFILE,fnm));
      warning( sWarn );
    }
    else {
      if (bVerbose==TRUE) printf("CP puckering Parameter Theta: %s\n", grpnm);
      writeTimePlot(opt2fn("-ot",NFILE,fnm), ePlotTHETA, grpnm, first_Frame);
    }
  }
  if (bPhi) {
  /* Write PhiXVG */
    if (bVerbose==TRUE) printf("CP puckering Parameter Phi: %s\n", grpnm);
    writeTimePlot(opt2fn("-op",NFILE,fnm), ePlotPHI, grpnm, first_Frame);
  }
  if (bQ) {
  /* Write QXVG */
    writeTimePlot(opt2fn("-oq",NFILE,fnm), ePlotQ, grpnm, first_Frame);
  }
  if (bThetaPhi) {
  /* TODO: Write Rama-Like Plot into seperate Funtion?*/
    /* The Title */
    sprintf(title, "CP Theta vs. Phi: %s", grpnm);
    out = xvgropen(opt2fn("-otp",NFILE,fnm), title,"\\f{Symbol}q\\f{} []", "\\f{Symbol}j\\f{} []");  // \f{Symbol} j \f{} :  = Symbol for phi
    xvgr_line_props(out,0,elNone,ecFrank);
    xvgr_view(out,0.2,0.2,0.8,0.8);
    xvgr_world(out,-180,-180,180,180);
    fprintf(out,"@    xaxis  tick on\n@    xaxis  tick major 60\n@    xaxis  tick minor 30\n");
    fprintf(out,"@    yaxis  tick on\n@    yaxis  tick major 60\n@    yaxis  tick minor 30\n");
    fprintf(out,"@ s0 symbol 2\n@ s0 symbol size 0.2\n@ s0 symbol fill 1\n");
    fprintf(out,"@TYPE XY\n");
    /*
          @with g0
          @    world -180, -180, 180, 180
          @    world xmin -180
          @    world xmax 180
          @    world ymin -180
          @    world ymax 180
          @    title "CP Theta vs. Phi: cp6"
          @    xaxis  label "\f{Symbol}q\f{} []"
          @    yaxis  label "\f{Symbol}j\f{} []"
          @    xaxis  tick major 60
          @    xaxis  tick minor ticks 1
          @    yaxis  tick major 60
          @    yaxis  tick minor ticks 1
          @    s0 type xy
          @    s0 symbol 9
          @    s0 symbol size 0.080000
          @    s0 line type 0
          @target G0.S0
          @type xy
    */
    current_Frame=first_Frame;
    while (current_Frame) {
      current_Ring=current_Frame->rings;
      while (current_Ring){
        fprintf(out, "%10.3f", current_Ring->fTheta>180 ? current_Ring->fTheta-360 : current_Ring->fTheta  );
        fprintf(out, "%10.3f\n", current_Ring->fPhi>180 ?   current_Ring->fPhi-360 :   current_Ring->fPhi  );
      }
    }
    fclose(out);
  }
  if (bPolar) {
  /* Write Polar Plot */
    writePolarPlot(opt2fn("-opol",NFILE,fnm), grpnm, first_Frame);
  }
  if (bThetaDist) {
    writeCpDistributionPlot (opt2fn("-odt",NFILE,fnm), ePlotTHETA, grpnm, first_Frame ) ;
  }
  if (bPhiDist) {
    writeCpDistributionPlot (opt2fn("-odp",NFILE,fnm), ePlotPHI, grpnm, first_Frame ) ;
  }

  print_warn_num();
  thanx(stderr);
  return 0;
}

